from typing import List

import pandas as pd


class JsonLDConverter:
    """
    Custom convertor from ORKG JSON to JSON-LD
    """

    _blank_node_counter = 1

    def __init__(self, host: str):
        self.ORKG_PREFIX = host
        if self.ORKG_PREFIX[-1] != "/":
            self.ORKG_PREFIX += "/"
        self.CLASS_PREFIX = f"{self.ORKG_PREFIX}class/"
        self.RESOURCE_PREFIX = f"{self.ORKG_PREFIX}resource/"
        self.PREDICATE_PREFIX = f"{self.ORKG_PREFIX}property/"

    def __handle_csvw_dataset(self, lol: List[List], context: dict) -> dict:
        # following CSVW @ https://www.w3.org/TR/2015/REC-tabular-data-model-20151217/
        if len(lol) == 0:
            return {}
        # create table node - the node that will be returned
        table_node = {"@type": [f"{self.CLASS_PREFIX}Table"], "label": "Table"}
        # create columns and assign them
        columns = lol[0]
        context["columns"] = f"{self.PREDICATE_PREFIX}CSVW_Columns"
        context["name"] = f"{self.PREDICATE_PREFIX}CSVW_Name"
        context["number"] = f"{self.PREDICATE_PREFIX}CSVW_Number"
        column_nodes = [
            {
                "@type": [f"{self.CLASS_PREFIX}Column"],
                "titles": [c],
                "number": i,
                "@id": self.get_blank_node_id(),
            }
            for i, c in enumerate(columns, start=1)
        ]
        table_node["columns"] = column_nodes
        # iterate over rows and create them
        row_nodes = []
        for i, r in enumerate(lol[1:], start=1):
            row = {
                "@type": [f"{self.CLASS_PREFIX}Row"],
                "titles": [f"Row {i}"],
                "cells": [],
                "number": i,
            }
            # run through cells
            for idx, c in enumerate(r):
                row["cells"].append(
                    {
                        "@type": [f"{self.CLASS_PREFIX}Cell"],
                        "value": str(c) if not pd.isna(c) else None,
                        "column": column_nodes[idx]["@id"],
                    }
                )
            # append row to rows
            row_nodes.append(row)
        context["rows"] = f"{self.PREDICATE_PREFIX}CSVW_Rows"
        context["cells"] = f"{self.PREDICATE_PREFIX}CSVW_Cells"
        context["value"] = f"{self.PREDICATE_PREFIX}CSVW_Value"
        context["titles"] = f"{self.PREDICATE_PREFIX}CSVW_Titles"
        context["column"] = f"{self.PREDICATE_PREFIX}CSVW_Column"
        table_node["rows"] = row_nodes
        return table_node

    def get_blank_node_id(self):
        """
        Creates an incremental blank node identifier for the json-ld structure
        :return: string in the form of _:n{number}
        """
        bn_id = f"_:n{self._blank_node_counter}"
        self._blank_node_counter = self._blank_node_counter + 1
        return bn_id

    def convert_2_jsonld(self, orkg_json: dict):
        """
        Convert an ORKG JSON format to an equivalent JSON-LD format
        :param orkg_json: The root ORKG JSON object
        :return: a json object in the format of JSON-LD
        """
        context = {}
        jsonld = {}
        if "resource" in orkg_json:
            orkg_json = orkg_json["resource"]
        # create dummy blank node
        jsonld["@id"] = self.get_blank_node_id()

        # handle label
        jsonld["label"] = orkg_json["name"]
        context["label"] = "http://www.w3.org/2000/01/rdf-schema#label"

        # handle classes
        jsonld["@type"] = [self.CLASS_PREFIX + c for c in orkg_json["classes"]]

        # handle values
        self.handle_values(orkg_json["values"], jsonld, context)

        # prepare final jsonld
        jsonld["@context"] = context
        return jsonld

    def handle_values(self, orkg_json: dict, jsonld: dict, context: dict):
        """
        Recursively handle the values part of the ORKG JSON format
        :param orkg_json: the ORKG object to work on
        :param jsonld: the resulting JSON-LD part
        :param context: the accompanying context for JSON-LD
        :return: Nothing, since it operates on the parameter dictionaries
        """
        # add predicates into context
        for prop, url in [(p, self.PREDICATE_PREFIX + p) for p in orkg_json.keys()]:
            context[prop] = url

        # handle each predicate
        for key, entities in orkg_json.items():
            jsonld[key] = []
            for entity in entities:
                jsonld_entity = {}
                # handle dataframe values
                if "@df" in entity:
                    jsonld_entity = self.__handle_csvw_dataset(entity["@df"], context)
                # handle @id
                if "@id" in entity:
                    if "@type" in entity:  # TODO: replace with match in python 3.10
                        entity_type = entity["@type"]
                        if entity_type == "resource":
                            jsonld_entity["@id"] = self.RESOURCE_PREFIX + entity["@id"]
                        elif entity_type == "class":
                            jsonld_entity["@id"] = self.CLASS_PREFIX + entity["@id"]
                        elif entity_type == "predicate":
                            jsonld_entity["@id"] = self.PREDICATE_PREFIX + entity["@id"]
                        else:
                            # FIXME: the literal value should be looked up and added here properly
                            jsonld_entity["@literal"] = entity["@id"]
                    else:
                        jsonld_entity["@id"] = self.RESOURCE_PREFIX + entity["@id"]

                # handle text
                if "text" in entity:
                    literal_value = entity["text"]
                    if "datatype" in entity:
                        literal_value = f"{literal_value}^^{jsonld_entity['datatype']}"
                    jsonld[key].append(literal_value)

                # handle label
                if "label" in entity:
                    jsonld_entity["@id"] = self.get_blank_node_id()
                    jsonld_entity["label"] = entity["label"]
                    if "classes" in entity:
                        jsonld_entity["@type"] = [
                            self.CLASS_PREFIX + c for c in entity["classes"]
                        ]

                    # handle values
                    if "values" in entity:
                        self.handle_values(entity["values"], jsonld_entity, context)
                # append newly created json-ld entity property
                if len(jsonld_entity) > 0:
                    jsonld[key].append(jsonld_entity)



from typing import List, Optional, Union

from orkg.client.harvesters import directory as directory_harvester
from orkg.client.harvesters import doi as doi_harvester
from orkg.common import OID
from orkg.out import OrkgResponse
from orkg.utils import NamespacedClient


class HarvestersClient(NamespacedClient):
    def doi_harvest(
        self, doi: str, orkg_rf: Union[str, OID], directory: Optional[str] = None
    ) -> OrkgResponse:
        """
        Harvests DOI data for a paper and add it to the ORKG.
        It works under the assumption that the paper contains some JSON-LD representation of its content
        If directory is provided, it is expected to have a `doi.json` file and other json files that are the contributions.
        If the `doi.json` doesn't exist and the `doi` parameter is present then the metadata is fetched from the DOI and the contributions from disk.
        :param doi: The DOI of the paper to harvest
        :param orkg_rf: The resource ID of the ORKG research field to add the harvested data to, or the string representation to be looked up (can raise errors)
        :param directory: The directory to read the specs from (optional)
        """
        return doi_harvester.harvest(
            orkg_client=self.client, doi=doi, orkg_rf=orkg_rf, directory=directory
        )

    def directory_harvest(
        self,
        directory: str,
        research_field: Union[str, OID],
        title: str,
        doi: Optional[str] = None,
        authors: Optional[List[Union[str, OID]]] = None,
        publication_year: Optional[int] = None,
        published_in: Optional[str] = None,
        **kwargs,
    ) -> OrkgResponse:
        """
        Harvests a directory of JSON-LD files created by the templates and add it to the ORKG.
        :param directory: The directory to read the specs from
        :param research_field: The resource ID of the ORKG research field to add the harvested data to, or the string representation to be looked up (can raise errors)
        :param title: The title of the paper
        :param doi: The DOI of the paper (optional)
        :param authors: The authors of the paper (optional)
        :param publication_year: The publication year of the paper (optional)
        :param published_in: The publication venue of the paper (optional)
        :param kwargs: Additional arguments to be passed to the paper object
        """

        return directory_harvester.harvest(
            orkg_client=self.client,
            directory=directory,
            research_field=research_field,
            title=title,
            doi=doi,
            authors=authors,
            publication_year=publication_year,
            published_in=published_in,
            **kwargs,
        )



from orkg.utils import NamespacedClient, query_params


class ComparisonsClient(NamespacedClient):
    @query_params("contribution_ids", "title", "description", "reference")
    def add(self, params=None):
        if "contribution_ids" not in params or "title" not in params:
            raise ValueError("at least contribution_ids and title should be provided")

        for contribution_id in params["contribution_ids"]:
            contribution_resource = self.client.resources.by_id(
                id=contribution_id
            ).content

            if "Contribution" not in contribution_resource["classes"]:
                raise ValueError(
                    "this ID is not a contribution: " + str(contribution_id)
                )

        description = params["description"] if "description" in params else ""
        reference = params["reference"] if "reference" in params else ""
        contribution_ids = ",".join(params["contribution_ids"])

        comparison_resp = self.client.resources.add(
            label=params["title"], classes=["Comparison"]
        )
        comparison_id = comparison_resp.content["id"]
        description_id = self.client.literals.add(label=description).content["id"]
        reference_id = self.client.literals.add(label=reference).content["id"]

        self.client.statements.add(
            subject_id=comparison_id,
            predicate_id="description",
            object_id=description_id,
        )
        self.client.statements.add(
            subject_id=comparison_id, predicate_id="reference", object_id=reference_id
        )

        self.client.json.save_json(
            resource_id=comparison_id,
            json_object={"url": "?contributions=" + contribution_ids},
        )

        print("Created successfully, comparison id: " + comparison_id)

        return self.client.wrap_response(comparison_resp)



from ast import literal_eval
from collections import defaultdict
from typing import Dict, List, Optional, Tuple, Union
from urllib.parse import parse_qs, urlparse

import pandas as pd

from orkg.out import OrkgResponse
from orkg.utils import NamespacedClient, simcomp_available


class ContributionsClient(NamespacedClient):
    @simcomp_available
    def similar(self, cont_id: str) -> OrkgResponse:
        """
        Get contributions which are similar to a specified contribution.
        Returns contribution ID, contribution label, paper ID, and similarity score.

        :param cont_id: contribution ID
        :return: OrkgResponse object of similar contributions
        """
        self.client.simcomp._append_slash = True
        response = self.client.simcomp.similar(cont_id).GET()
        return self.client.wrap_response(response=response)

    @simcomp_available
    def compare(
        self, contributions: List[str], response_hash: str = None
    ) -> OrkgResponse:
        """
        Get comparison by list of contribution IDs.

        :param contributions: list of contribution IDs
        :param response_hash: hashed URL of contribution (optional)
        :return: OrkgResponse object of the comparison
        """
        self.client.simcomp._append_slash = False
        params = f'?contributions={",".join(contributions)}'
        if response_hash is not None:
            params = f"{params}&response_hash={response_hash}"
        response = self.client.simcomp.compare.GET(params)
        return self.client.wrap_response(response=response)

    def _get_contributions_from_comparison(
        self, comparison_id: str
    ) -> Tuple[List[str], Optional[str], List[str]]:
        """
        Get contribution IDs of a comparison.

        :param comparison_id: ID of comparison
        :return:
            - single string of all contribution IDs from the comparison
            - string of hashed URL of contribution or None
            - list of comparison properties displayed in the UI
        """
        resource = self.client.resources.by_id(comparison_id).content
        if "Comparison" not in resource["classes"]:
            raise ValueError("This id is not for a comparison")
        response = self.client.json.get_json(resource_id=comparison_id).content
        query_params = parse_qs(urlparse(response["data"]["url"]).query)
        return (
            query_params["contributions"][0].split(","),
            query_params["response_hash"][0],
            query_params["properties"][0].split(",")
            if "properties" in query_params
            else [],
        )

    @simcomp_available
    def compare_dataframe(
        self,
        contributions: List[str] = None,
        comparison_id: str = None,
        like_ui=True,
        include_meta=False,
    ) -> Union[pd.DataFrame, Tuple[pd.DataFrame, pd.DataFrame]]:
        """
        Convert an ORKG comparison into a DataFrame.
        Optional: create a second DataFrame of contribution metadata.
        Included metadata (when present): author, doi, publication month, publication year, url, research field, venue

        :param contributions: list of contribution IDs from a comparison
        :param comparison_id: ID of a comparison
        :param like_ui: true/false whether to match comparison df to its UI representation
        :param include_meta: true/false whether to return metadata df
        :return:
            - comparison DataFrame
            - metadata DataFrame (optional)
        """
        # check if valid request
        if contributions is None and comparison_id is None:
            raise ValueError("either provide the contributions, or the comparison ID")
        # get OrkgResponse object
        response, ui_props = self._get_response(comparison_id, contributions)
        if not response.succeeded:
            return pd.DataFrame()
        content = response.content
        # extract necessary components for making dataframe
        (
            columns,
            contribution_ids_and_titles,
            contributions_list,
            property_lookup,
        ) = self._extract_components(content)
        # create table representation of the data
        data = content["data"]
        indices, rows = self._get_data_rows_and_indices(data, property_lookup)
        # create data frame from table rows and indices
        df = pd.DataFrame(rows, columns=columns, index=indices)
        if like_ui and len(ui_props) > 0:
            df = self._make_df_like_ui(df, property_lookup, ui_props)
        if include_meta:
            return df, self._create_metadata_df(
                columns, contribution_ids_and_titles, contributions_list
            )
        return df

    def _get_response(
        self, comparison_id: Optional[str], contributions: Optional[List]
    ) -> Tuple[OrkgResponse, List]:
        """
        Get an OrkgResponse object for a comparison by comparison ID or list of contributions IDs.

        :param comparison_id: ID of a comparison
        :param contributions: list of contribution IDs from a comparison
        :return:
            - OrkgResponse object of the comparison
            - list of comparison properties displayed in the UI
        """
        response_hash = None
        ui_props = []
        if comparison_id is not None:
            (
                contributions,
                response_hash,
                ui_props,
            ) = self._get_contributions_from_comparison(comparison_id)
        response = self.compare(
            contributions=contributions, response_hash=response_hash
        )
        return response, ui_props

    @staticmethod
    def _extract_components(
        content: Dict[str, Union[List, Dict]]
    ) -> Tuple[List[str], Dict[str, str], List[Dict], Dict[str, str]]:
        """
        Get information about contributions; create lookup dictionary for property IDs to labels; create column headings.

        :param content: OrkgResponse content
        :return:
            - list of headings for df columns
            - dictionary of contribution IDs to headings: Paper Title/Contribution N (Contribution ID)
            - list of contribution dictionaries containing contribution label, ID, paper ID, title, and year
            - dictionary of property IDs to property labels
        """
        contributions_list = content["contributions"]
        contribution_ids_and_titles = {
            contribution[
                "id"
            ]: f"{contribution['title']}/{contribution['contributionLabel']} ({contribution['id']})"
            for contribution in contributions_list
        }
        columns = [
            contribution_ids_and_titles[contribution["id"]]
            for contribution in contributions_list
        ]
        properties_list = content["properties"]
        property_lookup = {prop["id"]: prop["label"] for prop in properties_list}
        return columns, contribution_ids_and_titles, contributions_list, property_lookup

    def _get_data_rows_and_indices(
        self, data: Dict[str, List], property_lookup: Dict[str, str]
    ) -> Tuple[List, List]:
        """
        Convert comparison data into lists representing rows and indices corresponding to table representation of comparison.

        :param data: OrkgResponse.content['data']
        :param property_lookup: dictionary of property IDs to property labels
        :return:
            - list of comparison properties for the index
            - list of rows, with each row as a list of cell values
        """
        indices = []
        rows = []
        for prop_id, values in data.items():
            indices.append(property_lookup[prop_id])
            row = []
            for index, value in enumerate(values):
                if not value[0]:
                    row.append("")
                else:
                    if len(value) == 1:
                        row.append(self._get_cell_value(value[0]))
                    else:
                        cell = []
                        # Rather than parsing it in a list comprehension that might fail
                        # Then the whole list will be typed as strings, we parse it one item at a time
                        for v in value:
                            cell.append(self._get_cell_value(v))
                        row.append(cell)
            rows.append(row)
        return indices, rows

    @staticmethod
    def _get_cell_value(
        cell_dict: Dict[str, Union[str, List]]
    ) -> Union[str, float, int, bool]:
        """
        Convert cell value to appropriate Python datatype.

        :param cell_dict: dictionary of info for each cell.
        :return: cell value as a string, float, int, or bool.
        """
        try:
            # Try parse each value by itself
            value = literal_eval(cell_dict["label"].capitalize())
        except (SyntaxError, ValueError):
            # If it fails, that means it is a string
            value = cell_dict["label"]
        return value

    @staticmethod
    def _make_df_like_ui(
        df: pd.DataFrame, property_lookup: Dict[str, str], ui_props: List[str]
    ) -> pd.DataFrame:
        """
        Make the comparison DataFrame match the comparison as it appears in the UI by removing properties that aren't displayed and putting the rows in the same order.

        :param df: comparison DataFrame
        :param property_lookup: dictionary of property IDs to property labels
        :param ui_props: list of comparison properties displayed in the UI
        :return: comparison DataFrame
        """
        # remove props that are not displayed
        df = df.drop(
            [p_lbl for p_id, p_lbl in property_lookup.items() if p_id not in ui_props]
        )
        # order df rows by ui props order
        df = df.reindex([property_lookup[p] for p in ui_props if p in property_lookup])
        return df

    def _create_metadata_df(
        self,
        columns: List[str],
        contribution_ids_and_titles: Dict[str, str],
        contributions_list: List[Dict],
    ) -> pd.DataFrame:
        """
        Create a metadata DataFrame containing the following properties (when available):
            author, doi, publication month, publication year, url, research field, venue
        Column headings are identical to the comparison DataFrame.

        :param columns: list of headings for df columns
        :param contribution_ids_and_titles: dictionary of contribution IDs to headings: Paper Title/Contribution N (Contribution ID)
        :param contributions_list: list of contribution dictionaries containing contribution label, ID, paper ID, title, and year
        :return: metadata DataFrame
        """
        paper_set = set(
            [f"{contribution['paperId']}" for contribution in contributions_list]
        )
        comparison_meta_data = {}
        # create dict of meta properties and values
        for paper in paper_set:
            contribution_ids, paper_meta_data = self._get_paper_metadata(
                contributions_list, paper
            )
            # add paper meta info for each contribution to comparison meta info
            for contribution_id in contribution_ids:
                column_name = contribution_ids_and_titles[contribution_id]
                paper_dict = paper_meta_data.copy()
                comparison_meta_data[column_name] = paper_dict
                comparison_meta_data[column_name]["contribution id"] = contribution_id
        # Make dataframe with same column order as comparison df and replace missing fields with empty string
        df_meta = pd.DataFrame.from_dict(comparison_meta_data)[columns].fillna("")
        return df_meta

    def _get_paper_metadata(
        self, contributions_list: List[Dict], paper: str
    ) -> Tuple[List, Dict]:
        """
        Get metadata info for a paper.

        :param contributions_list: list of contribution dictionaries containing contribution label, ID, paper ID, title, and year
        :param paper: ID of the paper
        :return:
            - list of IDs of contributions belonging to this paper
            - dictionary of (available) paper metadata properties to values
        """
        # author, doi, publication month, publication year, url, research field, venue
        meta_property_ids = ["P27", "P26", "P28", "P29", "url", "P30", "HAS_VENUE"]
        paper_meta_dict_of_lists = defaultdict(list)
        paper_statements = self.client.statements.get_by_subject_unpaginated(
            subject_id=paper
        )
        for statement in paper_statements.content:
            if statement["predicate"]["id"] in meta_property_ids:
                pred = statement["predicate"]["label"]
                obj = statement["object"]["label"]
                paper_meta_dict_of_lists[pred].append(obj)
        # make dict values strings if only one and list of strings if two or more
        paper_meta_data = {
            k: v.pop() if len(v) == 1 else v
            for k, v in paper_meta_dict_of_lists.items()
        }
        paper_meta_data["title"] = paper_statements.content[0]["subject"]["label"]
        paper_meta_data["paper id"] = paper
        # get IDs of all contributions from this paper which are included in the comparison
        contribution_ids = [
            str(contribution["id"])
            for contribution in contributions_list
            if contribution["paperId"] == paper
        ]
        return contribution_ids, paper_meta_data



from typing import Optional, Union

import hammock
import requests
from requests import HTTPError

from orkg.common import Hosts
from orkg.out import OrkgResponse

from .classes import ClassesClient
from .comparisons import ComparisonsClient
from .contribution_comparisons import ContributionComparisonsClient
from .contributions import ContributionsClient
from .dummy import DummyClient
from .harvesters import HarvestersClient
from .json import JSONClient
from .literals import LiteralsClient
from .objects import ObjectsClient
from .papers import PapersClient
from .predicates import PredicatesClient
from .resources import ResourcesClient
from .statements import StatementsClient
from .stats import StatsClient
from .templates import TemplatesClient


class ORKG(object):
    """
    The base class for the ORKG client.
    Contains all the methods to interact with the ORKG API.
    """

    host: Optional[Union[str, Hosts]] = None
    simcomp_host: Optional[str] = None
    token: str = None

    def __init__(
        self,
        host: Optional[Union[str, Hosts]] = None,
        simcomp_host: Optional[str] = None,
        creds: Optional[tuple] = None,
        **kwargs,
    ):
        self._set_host(host)
        self._set_simcomp_host(simcomp_host)
        self.core = hammock.Hammock(self.host)
        if self.simcomp_available:
            self.simcomp = hammock.Hammock(self.simcomp_host)
        self.token = None
        if creds is not None and len(creds) == 2:
            self._authenticate(creds[0], creds[1])
        self.backend = self.core.api
        self.resources = ResourcesClient(self)
        self.predicates = PredicatesClient(self)
        self.classes = ClassesClient(self)
        self.literals = LiteralsClient(self)
        self.stats = StatsClient(self)
        self.statements = StatementsClient(self)
        self.papers = PapersClient(self)
        self.comparisons = ComparisonsClient(self)
        self.contributions = ContributionsClient(self)
        self.objects = ObjectsClient(self)
        self.dummy = DummyClient(self)
        self.json = JSONClient(self)
        self.templates = TemplatesClient(self)
        self.contribution_comparisons = ContributionComparisonsClient(self)
        self.harvesters = HarvestersClient(self)

    @property
    def simcomp_available(self):
        """
        Check whether the simcomp host is available or not
        :return: True if the simcomp host is available, False otherwise
        """
        return self.simcomp_host is not None

    def _set_host(self, host):
        """
        Set the host value of the backend based on the host passed by the user
        """
        if isinstance(host, Hosts):
            host = host.value
        if host is not None and not host.startswith("http"):
            if "host" not in host.lower():
                raise ValueError("host must begin with http or https")
            else:
                raise ValueError(
                    "the host name was not recognized "
                    "-- use Hosts.PRODUCTION, Hosts.SANDBOX, or Hosts.INCUBATING without quotes"
                )
        if host is not None and host[-1] == "/":
            host = host[:-1]
        self.host = host if host is not None else "https://sandbox.orkg.org"

    def _set_simcomp_host(self, simcomp_host):
        """
        Set the simcomp host
        :param simcomp_host: the simcomp host passed by the user
        """
        if simcomp_host is not None and not simcomp_host.startswith("http"):
            raise ValueError("simcomp host must begin with http or https")
        if simcomp_host is None and "orkg" in self.host:
            simcomp_host = self.host + "/simcomp"
        self.simcomp_host = simcomp_host

    def _authenticate(self, email, password):
        """
        Authenticate the user with the backend
        :param email: the email of the user
        :param password: the password of the user
        """
        data = {
            "username": email,
            "grant_type": "password",
            "client_id": "orkg-client",
            "password": password,
        }
        resp = self.core.oauth.token.POST(
            data=data, headers={"Authorization": "Basic b3JrZy1jbGllbnQ6c2VjcmV0"}
        )
        if resp.status_code == 200:
            self.token = resp.json()["access_token"]
        else:
            raise IOError(
                f"Please check the credentials provided!, got the error {resp.content}"
            )

    def ping(self) -> bool:
        """
        The ping function checks whether the backend is live or not.
        :returns: True if the backend is live, False otherwise
        """
        try:
            initial_response = requests.get(self.host, timeout=3, allow_redirects=True)
            initial_response.raise_for_status()
            return True
        except HTTPError:
            # TODO: log different status errors
            return False

    # FIXME: remove after system stabling
    def pagination_activated(self) -> bool:
        self.backend._append_slash = True
        return "pageable" in self.backend.resources.GET().json()

    def wrap_response(
        self,
        response=None,
        status_code: str = None,
        content: Union[list, dict] = None,
        url: str = None,
    ) -> OrkgResponse:
        """
        Wraps the response from the backend into a OrkgResponse object
        :param response: the response from the backend
        :param status_code: the status code of the response
        :param content: the content of the response
        :param url: the url of the response
        :return: OrkgResponse object
        """
        is_paged = self.pagination_activated()
        return OrkgResponse(
            client=self,
            response=response,
            status_code=status_code,
            content=content,
            url=url,
            paged=is_paged,
        )



import uuid
from typing import Dict, List, Union


def process_contribution(
    contribution_json: Union[Dict, List, str],
    resulting_object: Union[Dict, List],
    global_ids: Dict,
    context: Dict,
):
    if isinstance(contribution_json, List):
        for item in contribution_json:
            process_contribution(item, resulting_object, global_ids, context)
    elif isinstance(contribution_json, Dict):
        if isinstance(resulting_object, List):
            current_object = {}
            resulting_object.append(current_object)
            resulting_object = current_object
        for key, value in contribution_json.items():
            if key == "@id":
                global_ids[value] = "_" + str(uuid.uuid4())
                resulting_object["@temp"] = global_ids[value]
            elif key == "@type":
                resulting_object["classes"] = [
                    clazz.rsplit("/", 1)[-1] for clazz in contribution_json[key]
                ]
            elif key == "label":
                resulting_object["label"] = str(contribution_json[key])
            elif key == "@context":
                continue
            else:
                orkg_predicate_id = context[key].rsplit("/", 1)[-1]
                # other keys
                if "values" not in resulting_object:
                    resulting_object["values"] = {}
                if isinstance(value, List) and all(
                    isinstance(item, str) for item in value
                ):
                    resulting_object["values"][orkg_predicate_id] = [
                        {"text": str(item)} for item in value
                    ]
                elif isinstance(value, List) or isinstance(value, Dict):
                    predicate_values = []
                    resulting_object["values"][orkg_predicate_id] = predicate_values
                    process_contribution(value, predicate_values, global_ids, context)
                else:
                    # Check if the value starts with "_:"
                    # FIXME: Workaround because the descriptions aren't up to date!!
                    if key not in context:
                        orkg_predicate_id = f"CSVW_{key.title()}"
                    resulting_object["values"][orkg_predicate_id] = []
                    if isinstance(value, str) and value.startswith("_:"):
                        resulting_object["values"][orkg_predicate_id].append(
                            {"@id": global_ids[value]}
                        )
                    else:
                        resulting_object["values"][orkg_predicate_id].append(
                            {"text": str(value)}
                        )
        if "classes" in resulting_object and "label" not in resulting_object:
            resulting_object["label"] = ""



from typing import Dict, List, Union

from orkg.out import OrkgResponse
from orkg.utils import NamespacedClient, simcomp_available


class JSONClient(NamespacedClient):
    @simcomp_available
    def save_json(
        self, resource_id: str, json_object: Union[Dict, List]
    ) -> OrkgResponse:
        return self.client.wrap_response(
            self.client.simcomp.visualization.POST(
                json={"resourceId": resource_id, "jsonData": json_object}
            )
        )

    @simcomp_available
    def get_json(self, resource_id: str) -> OrkgResponse:
        params = f"?resourceId={resource_id}"
        return self.client.wrap_response(self.client.simcomp.visualization.GET(params))



from orkg.out import OrkgResponse, OrkgUnpaginatedResponse
from orkg.utils import NamespacedClient, dict_to_url_params, query_params


class PredicatesClient(NamespacedClient):
    def by_id(self, id) -> OrkgResponse:
        self.client.backend._append_slash = True
        response = self.client.backend.predicates(id).GET()
        return self.client.wrap_response(response)

    @query_params("q", "exact", "page", "size", "sort", "desc")
    def get(self, params=None) -> OrkgResponse:
        self.handle_sort_params(params)
        if len(params) > 0:
            self.client.backend._append_slash = False
            response = self.client.backend.predicates.GET(dict_to_url_params(params))
        else:
            self.client.backend._append_slash = True
            response = self.client.backend.predicates.GET()
        return self.client.wrap_response(response)

    @query_params("q", "exact", "page", "size", "sort", "desc")
    def get_unpaginated(
        self, params=None, start_page=0, end_page=-1
    ) -> OrkgUnpaginatedResponse:
        return self._call_pageable(
            self.get, args={}, params=params, start_page=start_page, end_page=end_page
        )

    @query_params("id", "label")
    def add(self, params=None) -> OrkgResponse:
        if len(params) == 0:
            raise ValueError("at least label must be provided")
        else:
            self.client.backend._append_slash = True
            response = self.client.backend.predicates.POST(
                json=params, headers=self.auth
            )
        return self.client.wrap_response(response)

    @query_params("id", "label")
    def find_or_add(self, params=None) -> OrkgResponse:
        if len(params) == 0:
            raise ValueError("at least label should be provided")
        else:
            if "id" in params:
                # check if a predicate with this id is there
                found = self.by_id(params["id"])
                if found.succeeded:
                    return found
            # check if a predicate with this label is there
            found = self.get(q=params["label"], exact=True, size=1)
            if found.succeeded:
                if isinstance(found.content, list) and len(found.content) > 0:
                    found.content = found.content[0]
                    return found
            # None found, what the hell! let's create a new one
            self.client.backend._append_slash = True
            response = self.client.backend.predicates.POST(
                json=params, headers=self.auth
            )
        return self.client.wrap_response(response)

    @query_params("label")
    def update(self, id, params=None) -> OrkgResponse:
        if len(params) == 0:
            raise ValueError("label must be provided")
        else:
            if not self.exists(id):
                raise ValueError("the provided id is not in the graph")
            self.client.backend._append_slash = True
            response = self.client.backend.predicates(id).PUT(
                json=params, headers=self.auth
            )
        return self.client.wrap_response(response)

    def exists(self, id) -> bool:
        return self.by_id(id).succeeded
